var fs = require('fs'),
	util = require('util'),
	osenv = require('osenv'),
	//os = require('os'),
	request = require('request'),
	FileCookieStore = require("tough-cookie-filestore"),
	clc = require('colors'),
	tar = require('tar'),
	fstream = require("fstream"),
	ncp = require("ncp"),
	path = require("path"),
	progress = require('request-progress'),
	sprintf = require('./sprintf'),
	msg = require('./msg');

var homepath = osenv.home();

clc.setTheme({
	silly: 'rainbow',
	input: 'grey',
	verbose: 'cyan',
	prompt: 'grey',
	info: 'green',
	data: 'grey',
	help: 'cyan',
	warn: 'yellow',
	debug: 'blue',
	error: 'red',
	empty: 'grey'
});

var prompt = require('./prompt');
var config=require('./../config');
var cookiepath=homepath+"/.esm-c.json";
var cookieStore=null;

var func={
	'JSONflatten':function(data) {
		var result = {};
		function recurse (cur, prop) {
			if (Object(cur) !== cur) {
				result[prop] = cur;
			} else if (Array.isArray(cur)) {
				for(var i=0, l=cur.length; i<l; i++)
					recurse(cur[i], prop + "[" + i + "]");
				if (l === 0)
					result[prop] = [];
			} else {
				var isEmpty = true;
				for (var p in cur) {
					isEmpty = false;
					//recurse(cur[p], prop ? prop+"."+p : p);
					recurse(cur[p], prop ? prop+"["+p+"]" : p);
				}
				if (isEmpty && prop)
					result[prop] = {};
			}
		}
		recurse(data, "");
		return result;
	},
	'fileExists':function(pth,cb){
		cb=cb||function(){};
		try {
			fs.lstat(pth,function(err,stats){
				if (stats) {
					if (stats.isFile()) {
						cb('file');
					} else if (stats.isDirectory()) {
						cb('dir');
					} else if (stats.isBlockDevice()) {
						cb('dev');
					} else if (stats.isCharacterDevice()) {
						cb('cdev');
					} else if (stats.isSymbolicLink()) {
						cb('link');
					} else {
						cb('other');
					}
				} else {
					cb(false);
				}
			});
		} catch (e) {
			console.log(e);
			cb(undefined);
		}
	},
	'createFile':function(pth,cnt,cb){
		cb=cb||function(){};
		fs.writeFile(pth, cnt, function(err) {
			if(err) {
				cb(undefined);
			} else {
				cb(true);
			}
		});
	},
	'readFile':function(pth,cb){
		cb=cb||function(){};
		fs.readFile(pth, function(err,dt) {
			if(err) {
				cb(undefined);
			} else {
				cb(dt);
			}
		});
	},
	'createDir':function(pth,cb){
		cb=cb||function(){};
		fs.mkdir(pth, function(err) {
			if(err) {
				cb(undefined);
			} else {
				cb(true);
			}
		});
	},
	'autoCreateDir':function(pth,cb){
		func.fileExists(pth,function(edir){
			if (edir && edir=='dir') {
				cb(true);
			} else if (edir === false) {
				func.createDir(pth,function(s){
					cb(s);
				});
			} else {
				cb(undefined);
			}
		});
	},
	createRecDir:function ensureDir(dir, mode, callback) {
		if (mode && typeof mode === 'function') {
			callback = mode;
			mode = null;
		}
		mode = mode || 0777 & (~process.umask());
		callback = callback || function () {};
		_ensureDir(dir, mode, callback);
	},
	createMultiDir:function(adir,cb) {
		var dir=adir||[];
		cb=cb||function(e){if (e) { console.log(clc.red('ERROR : cannot create folder "'+e+'"')); }};
		function mdr(i) {
			if (i<dir.length) {
				func.createRecDir(dir[i],function(err){
					if (!err) {
						mdr(i+1);
					} else {
						cb(dir[i]); //cannot create dir
						//console.log(clc.red('cannot create app dir'));
					}
				});
			} else {
				cb(false);
			}
		}
		mdr(0);
	},
	'currentUser':function(cb){
		func.request(func.apiurl+'me',function(error, response, body) {
			if (error) {
				cb('server connection failed');
				return;
			}
			var rslt= JSON.parse(body);
			if (rslt && rslt.dt && rslt.dt.userId) {
				cb(null,rslt.dt);
				//func.request(func.apiurl+'user/'+rslt.dt.userId,function(error2, response2, body2) {
				//	if (error2) {
				//		cb({err:'server connection failed'});
				//		return;
				//	}
				//	var rslt2= JSON.parse(body2);
				//	if (rslt2) {
				//		cb(rslt2);
				//	} else {
				//		cb({err:'can not get user detail'});
				//	}
				//});
			} else {
				cb(msg.not_logged_in);
			}
		});
	},
	'initCookie':function (cb) {
		function iCookie(){
			cookieStore=new FileCookieStore(cookiepath);
			var j = func.request.jar(cookieStore);
			func.request = func.request.defaults({ jar : j });
			cb();
		}
		if (cookieStore) {
			var j = func.request.jar(cookieStore);
			func.request = func.request.defaults({ jar : j });
			cb();
			return;
		}
		func.fileExists(cookiepath,function(ft){
			if (!ft) {
				func.createFile(cookiepath,'{}',function(s){
					if (s) {
						iCookie();
					} else {
						console.log(sprintf(msg.cannot_write_file,cookiepath));
						process.exit();
					}
				});
			} else if (ft=='file') {
				iCookie();
			} else {
				console.log(sprintf(msg.invalid_file_format,cookiepath));
				process.exit();
			}
		});
	},
	'arguments':function(l,s){
		return func.p.arguments[l]||func.p.options[l]||func.p.options[s]||null;
	},
	'repeat':function(s,n) {
		return (new Array(n+1).join(s));
	},
	'map':function(o,cb){
		cb=cb||function(v,k){};
		if (util.isArray(o)) {
			for(var i=0;i<o.length;i++){
				cb(o[i],i);
			}
		} else if(util.isObject(o)){
			for(var j in o){
				cb(o[j],j);
			}
		}
	},
	'keys':function(o){
		var ks=[];
		func.map(o,function(v,k){ks.push(k);});
		return ks;
	},
	'values':function(o){
		var vs=[];
		func.map(o,function(v,k){vs.push(v);});
		return vs;
	},
	'padLeft':function(s,l){
		return s.length<l?func.repeat(' ',l-s.length)+s:s;
	},
	'padRight':function(s,l){
		return s.length<l?s+func.repeat(' ',l-s.length):s;
	},
	'update':function(o1,o2){
		o1=o1||{};
        for(var i=1;i<arguments.length;i++){
             for(var ii in arguments[i]){
                 o1[ii]=arguments[i][ii];
             }    
        }
        return o1;
		//o1=o1||{}; o2=o2||o2;
		//for (var m in o2){
		//	o1[m]=o2[m];
		//}
		//return o1;
	},
	'formatOutput':function(o,space,addspace){
		space=space||1;
		addspace=addspace||2;
		if(util.isPrimitive(o)){
			console.log(func.repeat(' ',space)+o);
		} else if (util.isArray(o)) {
			func.map(o,function(v){
				func.formatOutput(v,space+addspace);
			});
		} else if (util.isObject(o)) {
			if (o.toString() && o.toString() != '[object Object]') {
				func.formatOutput(o.toString(),space+addspace);
			} else {
				var mlength=0;
				func.map(o,function(v,k){
					mlength=k.length>mlength?k.length:mlength;
				});
				func.map(o,function(v,k){
					if (util.isPrimitive(v)) {
						console.log(func.repeat(' ',space)+func.padRight(k,mlength)+':'+v);
					} else {
						console.log(func.repeat(' ',space)+func.padRight(k,mlength)+':');
						func.formatOutput(v,space+addspace);
					}
				});
			}
		}
	},
	deleteFolderRecursive:function(pth) {
		var files = [];
		if( fs.existsSync(pth) ) {
			if (fs.lstatSync(pth).isDirectory()) {
				files = fs.readdirSync(pth);
				files.forEach(function(file,index){
					var curPath = pth + "/" + file;
					if(fs.lstatSync(curPath).isDirectory()) { // recurse
						func.deleteFolderRecursive(curPath);
					} else { // delete file
						fs.unlinkSync(curPath);
					}
				});
				fs.rmdirSync(pth);
			} else {
				fs.unlinkSync(pth);
			}
		}
	},
	'srcStruct':function(pth,mtch){
		mtch=mtch||/\.js$/;
		var hpath=[];
		function struct(pth2){
				if (fs.lstatSync(pth2).isDirectory()) {
					files = fs.readdirSync(pth2);
					files.forEach(function(file,index){
						var curPath = pth2 + "/" + file;
						if(fs.lstatSync(curPath).isDirectory()) { // recurse
							struct(curPath);
						} else { // delete file
							if (curPath.match(mtch)) {
								hpath.push(curPath);
							}
						}
					});
					//fs.rmdirSync(pth2);
				} else {
					if (pth2.match(mtch)) {
						hpath.push(pth2);
					}
				}
		}
		if(fs.existsSync(pth)) {
			struct(pth);
		}
		return hpath;
	},
	spinner:function (message) {
		var spinnerMessage = message;
		this.start = function () {
			var self = this;
			var spinner = ['|','/','-','\\'];
			function play(arr, interval) {
				var len = arr.length, i = 0;
				interval = interval || 100;
				var drawTick = function () {
					var str = arr[i++ % len];
					process.stdout.write('\u001b[0G' + str + '\u001b[90m' + spinnerMessage + '\u001b[0m');
				};
				self.timer = setInterval(drawTick, interval);
			}
			var frames = spinner.map(function(c) {
				return util.format('  \u001b[96m%s ', c);
			});
			play(frames, 70);
		};
		this.message = function (message) {
			spinnerMessage = message;
		};
		this.stop = function () {
			process.stdout.write('\u001b[0G\u001b[2K');
			clearInterval(this.timer);
		};
	},
	'pack':function(cb){
		//console.log('pack dipanggil');
		cb=cb||function(){};
		var hsl=func.tmpdir+'/package.tar',
			hdir=func.tmpdir+'/package',
			dirDest;
		function onError(err) {
			console.error('An error occurred:', err);
		}
		function onEnd() {
			//console.log('pack selesai');
			//process.nextTick(function(){
				func.fileExists(hsl,function(r){
					if (r) {
						cb(hsl);
					} else {
						cb();
					}
				});
			//});
		}
		
		try {
			fs.unlinkSync(hsl);
		} catch(e) {}
		try {
			fs.mkdirSync(func.tmpdir);
		} catch(e){}
		
		func.deleteFolderRecursive(hdir);
		process.nextTick(function(){
			
			ncp(func.workingdir, hdir,{filter:function(name){
						return !(name.match(/node_modules\/?$/) ||
							name.match(/esm_modules\/?$/) ||
							name.match(/esm_libs\/?$/) ||
							name.match(/\.DS_Store$/)||
							name.match(/\.git/));
					}
				},function (err) {
					if (err) {
						console.log(msg.error,err);
						cb();
						return;
					}
					
					dirDest = fs.createWriteStream(hsl);
					
					var packer = tar.Pack({ noProprietary: true })
						.on('error', onError)
						.on('end', onEnd);
					
					dirDest.on('error', function(err) {
						//console.error("ERROR:",err);
						packer.read();
					});
					
					fstream.Reader({ path: hdir, type: "Directory" })
						.on('error', onError)
						.pipe(packer)
						.pipe(dirDest);
				});
		});
	},
	'pull':function(cb){
		if (!func.p.subcommand) {
			console.log(clc.red('ERROR : no package name'));
			return;
		}
		func.request(func.apiurl+'me',function(error, response, body) {
			if (error) { console.log(clc.red('ERROR : server connection failed')); return; }
			var r=JSON.parse(body);
			//console.log('r==>',r);
			if (!r.s) { console.log(clc.red('ERROR : '+r.e)); return; }
			var user=r.dt;
			func.request(func.apiurl+'package/'+func.p.subcommand,function(error, response, body) {
				if (error) { console.log(clc.red('ERROR : server connection failed')); return; }
				var r=JSON.parse(body);
				if (!r.s) { console.log(clc.red('ERROR : '+r.e)); return; }
				if (!r.dt) { console.log(clc.red('ERROR : package not found')); return; }
				if (r.dt.author!=user.userId) { console.log(clc.red('ERROR : not authorized')); return; }
				
				//var hsl=func.tmpdir+'/package.tar';
				//var hdir=func.tmpdir+'/package';
				var pkg=r.dt,
					fn=func.tmpdir+'/'+pkg.id+'@'+pkg.version+'.tar',
					dfn=func.tmpdir+'/'+pkg.id;
				try{ fs.unlinkSync(fn); } catch(e){ }
				var cd=path.resolve('./')
				func.deleteFolderRecursive(dfn);
				//console.log('dfn==>',dfn);
				dl(func.apiurl+'archive/package/'+pkg.id+'/'+pkg.version+'.tar',fn,pkg.id+'@'+pkg.version,function(e){
					if (e) {
						console.log(clc.red('ERROR : download error')); return; 
					}
					//console.log('download success');
					function onE(e){
						console.log(clc.red('ERROR:'+e.toString()));
					}
					var extractor = tar.Extract({path: dfn,strip: 0,})
						.on('error', onE)
						.on('end', function(){
							ncp(path.resolve(dfn+'/package'),func.workingdir,{stopOnErr:true},function(e){
								func.deleteFolderRecursive(dfn);
								if (e) {
									console.log(clc.red('ERROR:'+e.toString()));
									return;
								}
								console.log(clc.info('complete'));
							});
						});
					fs.createReadStream(fn)
						.on('error', onE)
						.pipe(extractor);
				});
			});
		});
	},
	//'install':function(cb){
	//	var pkg;
	//	try {
	//		pkg=require(func.esmpackage);
	//	} catch(e) {
	//		pkg=null;
	//	}
	//	
	//	function install(ids){
	//		func.request(func.apiurl+'deps/?id='+ids,function(error, response, body) {
	//			if (error) { console.log(clc.red('ERROR : server connection failed')); return; }
	//			var r=JSON.parse(body);
	//			if (!r.s) { console.log(clc.red('ERROR : '+r.e)); return; }
	//			if (!r.dt) { console.log(clc.red('ERROR : while get dependency list')); return; }
	//
	//			installPackages(r.dt,function(e){
	//				if (e) {
	//					console.log(clc.red('ERROR:'+e.toString()));
	//					return;
	//				}
	//				console.log(clc.info('complete'));
	//			});
	//		});
	//	}
	//	
	//	if (func.p.subcommand) {
	//		var ids=func.p.subcommand.split(',');
	//		
	//		//console.log('masuk ==> s',func.p.options)
	//		if(func.p.options && (func.p.options.s || func.p.options.save)){
	//			if(!pkg){
	//				console.log(clc.red('cannot read esm.json'));
	//				return;
	//			}
	//			pkg.dependencies=pkg.dependencies||[];
	//			pkg.dependencies.push.apply(pkg.dependencies,ids);
	//			func.createFile(func.esmpackage,JSON.stringify(pkg,true,'  '),function(rslt2){
	//				if (rslt2) {
	//					install(ids);
	//				} else {
	//					console.log(clc.red('cannot write esm.json file'));
	//				}
	//			});
	//			return;
	//		}
	//		install(ids);
	//		return;
	//	}
	//
	//	if(!pkg){
	//		console.log(clc.red('cannot read esm.json'));
	//		process.exit();
	//	}
	//	var ids=(pkg.dependencies||[]).join(',');
	//	if (!ids) {
	//		console.log(clc.empty('no dependencies'));
	//		return;
	//	}
	//	install(ids);
	//},
	'initCLI':function(cb){
		func.initCookie(cb);
	},
	'updateEsmJson':function(dt,cb){
		func.createFile(func.esmpackage,JSON.stringify(dt,true,'  '),function(rslt){
			if (rslt) {
				cb(true);
			} else {
				cb(false);
			}
		});
	}
};

module.exports=function(p){
	if (p) {
		func.p=p;
	}
	func.request=request;
	func.config=config;
	func.apiurl=config.apiurl;
	func.emobiqurl=config.emobiq;
	func.prompt=prompt;
	func.clc=clc;
	func.workingdir=process.cwd()+'/';
	func.esmpackage=func.workingdir+'esm.json';
	func.homepath=homepath;
	func.msg=msg;
	func.sprintf=sprintf;
	func.tmpdir=osenv.tmpdir();

	return func;
};

function _ensureDir(dir, mode, callback) {
  var existsFunction = fs.exists || path.exists;
  existsFunction(dir, function (exists) {
    if (exists) return callback(null);

    var current = path.resolve(dir);
    var parent = path.dirname(current);

    _ensureDir(parent, mode, function (err) {
      if (err) return callback(err);

      fs.mkdir(current, mode, function (err) {
        if (err) return callback(err);
        callback();
      });
    });
  });
}

function dl(uri,fln,label,done){
    process.stdout.write('downloading : '+label+'...');
     
    progress(request(uri), {
        throttle: 200,  // Throttle the progress event to 2000ms, defaults to 1000ms 
        delay: 100      // Only start to emit after 1000ms delay, defaults to 0ms 
    })
    .on('progress', function (state) {
		process.stdout.clearLine();
		process.stdout.cursorTo(0);
		process.stdout.write('receive ('+label+'):'+ state.received+'/'+state.total+' bytes ('+state.percent+'%)');
    })
    .on('error', function (err) {
        //console.log('err1');
        done(1);
    })
    .pipe(fs.createWriteStream(fln))
    .on('error', function (err) {
        //console.log('err2');
        done(1);
    })
    .on('close', function () {
		process.stdout.clearLine();
		process.stdout.cursorTo(0);
		var stats = fs.statSync(fln);
		var fileSizeInBytes = stats["size"];
		if (fileSizeInBytes < 50) {
			var s=fs.readFileSync(fln,'utf8'),
				so=null;
			try {
				so=JSON.parse(s);
			} catch(err) {
				so=null;
				//document.getElementById("demo").innerHTML = err.message;
			}
			if (so && !so.s) {
				fs.unlinkSync(fln);
				done(1);
				return;
			}
		}
		process.stdout.write('received done : '+label+'\n');
		//if (done) {
		done();
		//}
    });
}

function installPackages(ids,cb){
	if(!ids){ return; }
	ids=typeof ids == 'string'?[ids]:ids;
	
	function done(e) {
		if (e) { return cb(e); }
		if (ids.length) { return installPackages(ids,cb); }
		return cb();
	}
	var pid=ids.shift();
	
	if (!fs.existsSync('./esm_modules')) {
		fs.mkdirSync('./esm_modules');
	}
	
	if (fs.existsSync('./esm_modules/'+pid)) {
		done();
		return;
	}
	
	func.request(func.apiurl+'package/'+pid,function(error, response, body) {
		if (error) { done('server connection failed'); return; }
		var r=JSON.parse(body);
		if (!r.s) { done(r.e); return; }
		if (!r.dt) { done('package not found'); return; }
				
		var pkg=r.dt,
			fn=func.tmpdir+'/'+pkg.id+'@'+pkg.version+'.tar',
			dfn=func.tmpdir+'/'+pkg.id;
		
		try{ fs.unlinkSync(fn); } catch(e){ }
		func.deleteFolderRecursive(dfn);
		dl(func.apiurl+'archive/package/'+pkg.id+'/'+pkg.version+'.tar',fn,pkg.id+'@'+pkg.version,function(e){
			if (e) {
				//console.log('pth==>',func.apiurl+'archive/package/'+pkg.id+'/'+pkg.version+'.tar');
				done('download error');
				return;
			}
			function onE(e){
				done(e);
			}
			var extractor = tar.Extract({path: dfn,strip: 0,})
				.on('error', onE)
				.on('end', function(){
					ncp(path.resolve(dfn+'/package/dist'),'./esm_modules/'+pid,{stopOnErr:true},function(e){ //distribution only
						if (e) { done(e); return; }
						ncp(path.resolve(dfn+'/package/esm.json'),'./esm_modules/'+pid+'/esm.json',{stopOnErr:true},function(e){
							func.deleteFolderRecursive(dfn);
							if (e) { done(e); return; }
							done();
						});
					});
				});
			fs.createReadStream(fn)
				.on('error', onE)
				.pipe(extractor);
		});
	});
}