var readline = require('readline');
var Mute = require('mute-stream');
var rl;

function read (opts, cb) {
	if (opts.num) {
		throw new Error('read() no longer accepts a char number limit');
	}
	
	if (typeof opts.default !== 'undefined' &&
		typeof opts.default !== 'string' &&
		typeof opts.default !== 'number') {
		throw new Error('default value must be string or number');
	}
	
	var input = opts.input || process.stdin;
	var output = opts.output || process.stdout;
	var prompt = (opts.prompt || '') + ' ';
	var silent = opts.silent;
	var editDef = false;
	var timeout = opts.timeout;
	
	var def = opts.default || '';
	if (def) {
		if (silent) {
			prompt += '(<default hidden>) ';
		} else if (opts.edit) {
			editDef = true;
		} else {
			prompt += '(' + def + ') ';
		}
	}
	var terminal = !!(opts.terminal || output.isTTY);
	
	var m = new Mute({ replace: opts.replace, prompt: prompt });
	m.pipe(output, {end: false});
	output = m;
	var rlOpts = { input: input, output: output, terminal: terminal };
	
	if (process.version.match(/^v0\.6/)) {
		rl = readline.createInterface(rlOpts.input, rlOpts.output);
	} else {
		rl = readline.createInterface(rlOpts);
	}
	
	
	output.unmute();
	rl.setPrompt(prompt);
	rl.prompt();
	if (silent) {
		output.mute();
	} else if (editDef) {
		rl.line = def;
		rl.cursor = def.length;
		rl._refreshLine();
	}
	
	var called = false;
	rl.on('line', onLine);
	rl.on('error', onError);
	
	rl.on('SIGINT', function () {
		rl.close();
		onError(new Error('canceled'));
	});
	
	var timer;
	if (timeout) {
		timer = setTimeout(function () {
			onError(new Error('timed out'));
		}, timeout);
	}
	
	function done () {
		called = true;
		rl.close();
	  
		if (process.version.match(/^v0\.6/)) {
			rl.input.removeAllListeners('data');
			rl.input.removeAllListeners('keypress');
			rl.input.pause();
		}
	  
		clearTimeout(timer);
		output.mute();
		output.end();
	}
	
	function onError (er) {
		if (called) return;
		done();
		return cb(er);
	}
	
	function onLine (line) {
		if (called) return;
		if (silent && terminal) {
			output.unmute();
			output.write('\r\n');
		}
		done();
		// truncate the \n at the end.
		line = line.replace(/\r?\n$/, '');
		var isDefault = !!(editDef && line === def);
		if (def && !line) {
			isDefault = true;
			line = def;
		}
		cb(null, line, isDefault);
	}
}

module.exports=function(arg,cb){
	arg=arg||{};
	var akey=[];
	var rslt={};
	var ikey=0;
	
	for(var i in arg){
		akey.push(i);
		arg[i].prompt=arg[i].prompt||('  '+i+':');
	}

	function reset(c) {
		var a='prompt silent replace timeout default edit terminal input output'.split(' ');
		var x={};
		for(var i=0;i<a.length;i++){
			if (c[a[i]]!==undefined) {
				x[a[i]]=c[a[i]];
			}
		}
		return x;
	}

	function _read(_i) {
		var ckey=akey[_i];
		var carg=arg[ckey];
		read(reset(carg),function(e,r){
			if (e) {
				cb(e);
			} else {
				if (!r && carg.required) {
					console.log(carg.message||ckey+' is required');
					_read(_i);
				} else {
					if (carg.pattern && !(r+'').match(carg.pattern)) {
						console.log(carg.message||ckey+' invalid value');
						_read(_i);
					} else {
						rslt[ckey]=r;
						if (_i+1 >= akey.length) {
							cb(false,rslt);
						} else {
							_read(_i+1);
						}
					}
				}
			}
		});
	}

	if (akey.length) {
		_read(0);
	} else {
		cb(false,{});
	}
};